//! MAGICC components for RSCM
//!
//! This crate scaffolds the structure for future MAGICC component implementations

pub mod carbon;
pub mod chemistry;
pub mod climate;
pub mod forcing;
pub mod python;
