# This .pyi file and the other .pyi files within this module are handwritten.
# See https://github.com/PyO3/pyo3/issues/2454 for more information about the ongoing
# implementation of the automatic generation of .pyi files from rust

__version__: str
